# mypy: ignore-errors
#
# PEP 695 isn't yet supported in mypy. (April 4, 2024)
from dataclasses import dataclass
from typing import Annotated, Any, Literal, NewType

import pytest
from helptext_utils import get_helptext_with_checks
from pydantic import BaseModel

import tyro
from tyro.conf._markers import OmitArgPrefixes
from tyro.constructors._primitive_spec import UnsupportedTypeAnnotationError


def test_simple_generic() -> None:
    @dataclass(frozen=True)
    class Container[T]:
        a: T
        b: T

    assert tyro.cli(Container[int], args="--a 1 --b 2".split(" ")) == Container(1, 2)


type X = int
type Y = list[int]
type Z = Inner[int]


@dataclass(frozen=True)
class Inner[T]:
    a: T
    b: T


def test_generic_with_type_statement_0() -> None:
    @dataclass(frozen=True)
    class Container[T]:
        a: T
        b: T

    assert tyro.cli(Container[X], args="--a 1 --b 2".split(" ")) == Container(1, 2)


def test_generic_with_type_statement_1() -> None:
    @dataclass(frozen=True)
    class Container[T]:
        a: tuple[X, ...]
        b: T

    assert tyro.cli(Container[Y], args="--a 1 --b 2".split(" ")) == Container((1,), [2])


def test_generic_with_type_statement_2() -> None:
    @dataclass(frozen=True)
    class Container[T]:
        a: Z

    assert tyro.cli(Container[Y], args="--a.a 1 --a.b 2".split(" ")) == Container(
        Inner(1, 2)
    )


type AnnotatedBasic = Annotated[int, tyro.conf.arg(name="basic")]


def test_annotated_alias() -> None:
    @dataclass(frozen=True)
    class Container:
        a: AnnotatedBasic

    assert tyro.cli(Container, args="--basic 1".split(" ")) == Container(1)


type TT[T] = Annotated[T, tyro.conf.arg(name="", constructor=lambda: True)]


def test_pep695_generic_alias() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: TT[bool]

    assert tyro.cli(Config, args=[]) == Config(arg=True)


type Renamed[T] = Annotated[T, tyro.conf.arg(name="renamed")]


def test_pep695_generic_alias_rename() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: Renamed[bool]

    assert tyro.cli(Config, args=["--renamed", "True"]) == Config(arg=True)


def test_pep695_generic_alias_rename_override() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: Annotated[Renamed[bool], tyro.conf.arg(name="renamed2")]

    assert tyro.cli(Config, args=["--renamed2", "True"]) == Config(arg=True)


type RenamedTwice[T] = Renamed[Renamed[T]]


def test_pep695_generic_alias_rename_twice() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: RenamedTwice[bool]

    assert tyro.cli(Config, args=["--renamed", "True"]) == Config(arg=True)


type SomeUnion = bool | int
type RenamedThreeTimes[T] = Renamed[Renamed[Renamed[T]]]
type SomeUnionRenamed = RenamedThreeTimes[SomeUnion]


def test_pep695_generic_alias_rename_three_times() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: Annotated[SomeUnionRenamed, tyro.conf.arg(name="renamed_override")]

    assert tyro.cli(Config, args=["--renamed-override", "True"]) == Config(arg=True)


type RecursiveList[T] = T | list[RecursiveList[T]]


def test_pep695_recursive_types() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: RecursiveList[str]

    # Recursive types are still unsupported.
    with pytest.raises((UnsupportedTypeAnnotationError, SystemExit)):
        tyro.cli(Config, args=["--arg", "True"])


def test_pep695_recursive_types_custom_constructor() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: Annotated[RecursiveList[str], tyro.conf.arg(constructor=str)]

    assert tyro.cli(Config, args=["--arg", "True"]) == Config(arg="True")


type UnprefixedSubcommandPair[T1, T2, T3] = (
    Annotated[OmitArgPrefixes[T1], tyro.conf.subcommand(prefix_name=False)]  # type: ignore
    | Annotated[T2, tyro.conf.subcommand(prefix_name=False)]  # type: ignore
    | Annotated[T3, tyro.conf.subcommand(prefix_name=True)]
)
type IntContainer = Inner[int]
type IntContainerIntermediate = Inner[int]
type IntContainer2 = IntContainerIntermediate
type StrContainer = Inner[str]


def test_pep695_alias_subcommand() -> None:
    """Adapted from: https://github.com/brentyi/tyro/issues/177"""

    @dataclass(frozen=True)
    class Config:
        arg: UnprefixedSubcommandPair[IntContainer, IntContainer2, StrContainer]

    assert tyro.cli(Config, args=["int-container", "--a", "3", "--b", "5"]) == Config(
        Inner(3, 5)
    )
    assert tyro.cli(
        Config, args=["int-container2", "--arg.a", "3", "--arg.b", "5"]
    ) == Config(Inner(3, 5))
    assert tyro.cli(
        Config, args=["arg:str-container", "--arg.a", "3", "--arg.b", "5"]
    ) == Config(Inner("3", "5"))


type Int0 = int
Int1 = NewType("Int1", Int0)
type Int2 = Int1
Int3 = NewType("Int3", Int2)
type Int4 = Int3
Int5 = NewType("Int5", Int4)
type Int6 = Int5
Int7 = NewType("Int7", Int6)


def test_pep695_new_type_alias() -> None:
    def main(arg: list[Int7], /) -> Any:
        return arg

    assert tyro.cli(main, args=["1", "2"]) == [1, 2]


def test_generic_config() -> None:
    @dataclass(frozen=True)
    class Container[T]:
        a: Inner[T]

    assert tyro.cli(
        Container[bool],
        args="--a.a True --a.b False".split(" "),
        config=(tyro.conf.FlagConversionOff,),
    ) == Container(Inner(True, False))


def test_generic_config_subcommand() -> None:
    @dataclass(frozen=True)
    class Container[T]:
        a: T

    assert tyro.cli(
        Container[Container[bool] | Container[str]],
        args="a:container-bool --a.a True".split(" "),
        default=Container(Container(a="30")),
        config=(tyro.conf.FlagConversionOff,),
    ) == Container(Container(True))

    assert tyro.cli(
        Container[Container[bool] | Container[str]],
        args=[],
        default=Container(Container(a="30")),
        config=(tyro.conf.FlagConversionOff,),
    ) == Container(Container("30"))

    assert tyro.cli(
        Container[Container[bool] | Container[str]],
        args=[],
        default=Container(Container(a=False)),
        config=(tyro.conf.FlagConversionOff,),
    ) == Container(Container(False))


def test_generic_config_subcommand2() -> None:
    @dataclass(frozen=True)
    class Container[T]:
        a: tyro.conf.OmitSubcommandPrefixes[T]

    assert tyro.cli(
        Container[Container[bool] | Container[str]],
        args="container-bool --a True".split(" "),
    ) == Container(Container(True))


def test_generic_config_subcommand3() -> None:
    @dataclass(frozen=True)
    class Container[T]:
        a: T

    assert tyro.cli(
        Container[Container[bool] | Container[str]],
        args=[],
        default=Container(Container(a=True)),
        config=(tyro.conf.OmitSubcommandPrefixes,),
    ) == Container(Container(True))


def test_generic_config_subcommand4() -> None:
    @dataclass(frozen=True)
    class Container[T]:
        a: T

    assert tyro.cli(
        Container[Container[bool] | Container[Literal["1", "2"]]],
        args="container-literal-1-2 --a 2".split(" "),
        config=(tyro.conf.OmitSubcommandPrefixes,),
    ) == Container(Container("2"))
    assert tyro.cli(
        Container[Container[bool] | Container[Literal["1", "2"]]],
        args=[],
        default=Container(Container(a=True)),
        config=(tyro.conf.OmitSubcommandPrefixes,),
    ) == Container(Container(True))


def test_generic_config_subcommand_matching_nested() -> None:
    @dataclass(frozen=True)
    class Container[T]:
        a: T

    assert "default: a:container-bool" in get_helptext_with_checks(
        Container[Container[bool] | Container[str]],
        default=Container(Container(a=False)),
    )
    assert "default: a:container-str" in get_helptext_with_checks(
        Container[Container[bool] | Container[str]],
        default=Container(Container(a="False")),
    )
    assert "default: a:container-literal-1-2" in get_helptext_with_checks(
        Container[Container[Literal[1, 2]] | Container[str]],
        default=Container(Container(a=1)),
    )
    assert "default: a:container-str" in get_helptext_with_checks(
        Container[Container[Literal[1, 2]] | Container[str]],
        default=Container(Container(a="1")),
    )

    assert tyro.cli(
        Container[Container[Literal["1", "2"]] | Container[bool]],
        args=[],
        default=Container(Container(a="1")),
        config=(tyro.conf.OmitSubcommandPrefixes,),
    ) == Container(Container("1"))


def test_generic_config_subcommand_matching_dict() -> None:
    @dataclass(frozen=True)
    class Container[T]:
        a: T

    assert "default: container-dict-str-str" in get_helptext_with_checks(
        Container[dict[str, int]] | Container[dict[str, str]],  # type: ignore
        default=Container({"a": "text"}),
    )
    assert "default: container-dict-str-int" in get_helptext_with_checks(
        Container[dict[str, int]] | Container[dict[str, str]],  # type: ignore
        default=Container({"a": 5}),
    )


def test_generic_config_subcommand_matching_tuple() -> None:
    @dataclass(frozen=True)
    class Container[T]:
        a: T

    assert "default: container-tuple-str-str" in get_helptext_with_checks(
        Container[tuple[str, int]] | Container[tuple[str, str]],  # type: ignore
        default=Container(("a", "text")),
    )
    assert "default: container-tuple-str-int" in get_helptext_with_checks(
        Container[tuple[str, int]] | Container[tuple[str, str]],  # type: ignore
        default=Container(("a", 5)),
    )


def test_generic_config_subcommand_matching_tuple_variable() -> None:
    @dataclass(frozen=True)
    class Container[T]:
        a: T

    assert "default: container-tuple-str-ellipsis" in get_helptext_with_checks(
        Container[tuple[str, ...]] | Container[tuple[int, ...]],  # type: ignore
        default=Container(("a", "text")),
    )
    assert "default: container-tuple-int-ellipsis" in get_helptext_with_checks(
        Container[tuple[str, ...]] | Container[tuple[int, ...]],  # type: ignore
        default=Container((1, 2, 3)),
    )


def test_deeply_inherited_init() -> None:
    class AConfig(BaseModel):
        a: int

    class AModel[TContainsAConfig: AConfig]:
        def __init__(self, config: TContainsAConfig):
            self.config = config

    class ABConfig(AConfig):
        b: int

    class ABModel[TContainsABConfig: ABConfig](AModel[TContainsABConfig]):
        pass

    class ABCConfig(ABConfig):
        c: int

    class ABCModel(ABModel[ABCConfig]):
        pass

    def a(model: ABCModel):
        print(model.config)

    def b(model: ABModel[ABConfig]):
        print(model.config)

    def c(model: ABModel[ABCConfig]):
        print(model.config)

    assert "--model.config.a" in get_helptext_with_checks(a)
    assert "--model.config.b" in get_helptext_with_checks(a)
    assert "--model.config.c" in get_helptext_with_checks(a)

    assert "--model.config.a" in get_helptext_with_checks(b)
    assert "--model.config.b" in get_helptext_with_checks(b)
    assert "--model.config.c" not in get_helptext_with_checks(b)

    assert "--model.config.a" in get_helptext_with_checks(c)
    assert "--model.config.b" in get_helptext_with_checks(c)
    assert "--model.config.c" in get_helptext_with_checks(c)


def test_bad_orig_bases() -> None:
    @dataclass(frozen=True)
    class A[T]:
        a: T

    @dataclass(frozen=True)
    class B(A[int | str | bool]):
        x: int

    @dataclass(frozen=True)
    class C(A[int | str | bool]):
        a: str

    @dataclass(frozen=True)
    class D(B, C): ...  # , C): ...

    assert "--a" in get_helptext_with_checks(D)
    assert "STR" in get_helptext_with_checks(D)
    assert "INT|STR" not in get_helptext_with_checks(D)


type ConstrainedTuple[T] = tuple[T] | tuple[T, T]


def test_generic_type_alias_union() -> None:
    """Test that unions of generic type aliases resolve type parameters correctly.

    Regression test for bug where ConstrainedTuple[int] | ConstrainedTuple[bool]
    would incorrectly resolve to typing.tuple[int]| tuple[int, int]| tuple[T]| tuple[T, T]
    instead of typing.tuple[int]| tuple[int, int]| tuple[bool]| tuple[bool, bool].
    """

    def main(arg: ConstrainedTuple[int] | ConstrainedTuple[bool]) -> Any:
        return arg

    # Test with int tuple of length 1
    assert tyro.cli(main, args=["--arg", "1"]) == (1,)

    # Test with int tuple of length 2
    assert tyro.cli(main, args=["--arg", "1", "2"]) == (1, 2)

    # Test with bool tuple of length 1
    assert tyro.cli(main, args=["--arg", "True"]) == (True,)

    # Test with bool tuple of length 2
    assert tyro.cli(main, args=["--arg", "False", "True"]) == (False, True)


def test_generic_type_alias_individual_resolution() -> None:
    """Test that individual generic type aliases resolve correctly.

    This verifies that the type parameter resolution works correctly for
    individual generic type aliases before testing unions.
    """
    from tyro._resolver import TypeParamResolver

    # Test individual resolution
    alias_int = ConstrainedTuple[int]
    alias_bool = ConstrainedTuple[bool]

    resolved_int = TypeParamResolver.resolve_params_and_aliases(alias_int)  # type: ignore
    resolved_bool = TypeParamResolver.resolve_params_and_aliases(alias_bool)  # type: ignore

    # Check that the resolved types are what we expect

    assert resolved_int == tuple[int] | tuple[int, int]
    assert resolved_bool == tuple[bool] | tuple[bool, bool]


def test_generic_type_alias_union_resolution() -> None:
    """Test that unions of generic type aliases resolve correctly at the type level.

    This tests the internal type resolution mechanism that was buggy.
    """

    from tyro._resolver import TypeParamResolver

    # Test union resolution
    union_type = ConstrainedTuple[int] | ConstrainedTuple[bool]
    resolved_union = TypeParamResolver.resolve_params_and_aliases(union_type)  # type: ignore

    # The resolved union should contain all four tuple types
    expected = tuple[int] | tuple[int, int] | tuple[bool] | tuple[bool, bool]
    assert resolved_union == expected

    # Verify that no unresolved TypeVars remain (this was the bug)
    # The resolved type should not contain any raw TypeVar instances
    import typing
    from typing import get_args

    def contains_typevar(typ) -> bool:
        """Recursively check if a type contains any TypeVar instances."""
        if isinstance(typ, typing.TypeVar):
            return True
        for arg in get_args(typ):
            if contains_typevar(arg):
                return True
        return False

    assert not contains_typevar(resolved_union), (
        f"Resolved type contains unresolved TypeVars: {resolved_union}"
    )


def test_generics_inherited() -> None:
    """Test that generics in inherited classes work correctly.

    Adapted from: https://github.com/brentyi/tyro/issues/327
    """

    @dataclass(frozen=True)
    class Point3[ScalarType: (int, float)]:
        x: ScalarType
        y: ScalarType
        z: ScalarType
        frame_id: str

    @dataclass(frozen=True)
    class Triangle:
        a: Point3[float] = Point3(0.0, 0.0, 0.0, "frame")
        b: Point3[float] = Point3(0.0, 0.0, 0.0, "frame")
        c: Point3[float] = Point3(0.0, 0.0, 0.0, "frame")

    @dataclass(frozen=True)
    class Args[ShapeType]:
        shape: ShapeType

    @dataclass(frozen=True)
    class SubArgs[ShapeType](Args[ShapeType]):
        some_int: int = 1

    assert tyro.cli(
        SubArgs[Triangle], args=["--some-int", "2", "--shape.a.x", "5.0"]
    ) == SubArgs(
        shape=Triangle(
            Point3(5.0, 0.0, 0.0, "frame"),
            Point3(0.0, 0.0, 0.0, "frame"),
            Point3(0.0, 0.0, 0.0, "frame"),
        ),
        some_int=2,
    )


def test_generics_inherited_deep() -> None:
    """Test that generics in inherited classes work correctly.

    Adapted from: https://github.com/brentyi/tyro/issues/327
    """

    @dataclass(frozen=True)
    class Point3[ScalarType: (int, float)]:
        x: ScalarType
        y: ScalarType
        z: ScalarType
        frame_id: str

    @dataclass(frozen=True)
    class Triangle:
        a: Point3[float] = Point3(0.0, 0.0, 0.0, "frame")
        b: Point3[float] = Point3(0.0, 0.0, 0.0, "frame")
        c: Point3[float] = Point3(0.0, 0.0, 0.0, "frame")

    @dataclass(frozen=True)
    class Args[ShapeType]:
        shape: ShapeType

    @dataclass(frozen=True)
    class SubArgs[ShapeType1](Args[ShapeType1]):
        some_int1: int = 1

    @dataclass(frozen=True)
    class SubSubArgs[ShapeType2](SubArgs[ShapeType2]):
        some_int2: int = 1

    @dataclass(frozen=True)
    class SubSubSubArgs[ShapeType3](SubSubArgs[ShapeType3]):
        some_int3: int = 1

    assert tyro.cli(
        SubSubSubArgs[Triangle], args=["--some-int2", "2", "--shape.a.x", "5.0"]
    ) == SubSubSubArgs(
        shape=Triangle(
            Point3(5.0, 0.0, 0.0, "frame"),
            Point3(0.0, 0.0, 0.0, "frame"),
            Point3(0.0, 0.0, 0.0, "frame"),
        ),
        some_int2=2,
    )



"""Tests for tyro.MISSING."""

import contextlib
import dataclasses
import io
from typing import Tuple, Union

import pytest

import tyro


def test_missing() -> None:
    def main(a: int = 5, b: int = tyro.MISSING, c: int = 3) -> Tuple[int, int, int]:
        return a, b, c

    target = io.StringIO()
    with pytest.raises(SystemExit), contextlib.redirect_stderr(target):
        tyro.cli(main, args=[])
    message = target.getvalue()
    assert "Required options" in message
    assert "Argument helptext" in message or "Missing from" in message
    assert "--a INT" not in message
    assert "--b INT" in message
    assert "(required)" in message
    assert tyro.cli(main, args=["--b", "7"]) == (5, 7, 3)


def test_missing_dataclass() -> None:
    @dataclasses.dataclass
    class Args2:
        a: int = 5
        b: int = tyro.MISSING
        c: int = 3

    with pytest.raises(SystemExit):
        tyro.cli(Args2, args=[])
    assert tyro.cli(Args2, args=["--b", "7"]) == Args2(5, 7, 3)


def test_missing_default() -> None:
    @dataclasses.dataclass
    class Args2:
        a: int
        b: int
        c: int

    with pytest.raises(SystemExit):
        tyro.cli(
            Args2,
            args=[],
            default=Args2(5, tyro.MISSING, 3),
        )
    assert tyro.cli(
        Args2,
        args=["--b", "7"],
        default=Args2(5, tyro.MISSING, 3),
    ) == Args2(5, 7, 3)


def test_missing_nested_default() -> None:
    @dataclasses.dataclass
    class Child:
        a: int = 5
        b: int = 3
        c: int = 7

    @dataclasses.dataclass
    class Parent:
        child: Child

    with pytest.raises(SystemExit):
        tyro.cli(
            Parent,
            args=[],
            default=Parent(child=tyro.MISSING),
        )
    assert tyro.cli(
        Parent,
        args=["--child.a", "5", "--child.b", "7", "--child.c", "3"],
        default=Parent(child=tyro.MISSING),
    ) == Parent(Child(5, 7, 3))


def test_missing_in_dataclass_def() -> None:
    @dataclasses.dataclass
    class Child:
        a: int = 5
        b: int = 3
        c: int = 7

    @dataclasses.dataclass
    class Parent:
        child: Child = tyro.MISSING

    assert tyro.cli(
        Parent, args=["--child.a", "5", "--child.b", "7", "--child.c", "3"]
    ) == Parent(Child(5, 7, 3))

    with pytest.raises(SystemExit):
        tyro.cli(Parent, args=[], default=tyro.MISSING)

    # tyro.MISSING in dataclass definition should propagate to child fields, which makes all arguments required.
    with pytest.raises(SystemExit):
        tyro.cli(Parent, args=[])
    with pytest.raises(SystemExit):
        tyro.cli(Parent, args=["--child.a", "5"])


def test_missing_in_tuple() -> None:
    @dataclasses.dataclass
    class Child:
        a: int = 5
        b: int = 3
        c: int = 7

    @dataclasses.dataclass
    class Parent:
        child: Tuple[Child, Child] = tyro.MISSING

    with pytest.raises(SystemExit):
        tyro.cli(Parent, args=[])

    assert tyro.cli(
        Parent,
        args=[
            "--child.0.a",
            "5",
            "--child.0.b",
            "7",
            "--child.0.c",
            "3",
            "--child.1.a",
            "5",
            "--child.1.b",
            "7",
            "--child.1.c",
            "3",
        ],
    ) == Parent((Child(5, 7, 3), Child(5, 7, 3)))


def test_missing_in_tuple_pair() -> None:
    @dataclasses.dataclass
    class Child:
        a: int = 5
        b: int = 3
        c: int = 7

    @dataclasses.dataclass
    class Parent:
        child: Tuple[Child, Child] = (tyro.MISSING, tyro.MISSING)

    with pytest.raises(SystemExit):
        tyro.cli(Parent, args=[])

    assert tyro.cli(
        Parent,
        args=[
            "--child.0.a",
            "5",
            "--child.0.b",
            "7",
            "--child.0.c",
            "3",
            "--child.1.a",
            "5",
            "--child.1.b",
            "7",
            "--child.1.c",
            "3",
        ],
    ) == Parent((Child(5, 7, 3), Child(5, 7, 3)))


def test_missing_in_tuple_pair_subcommands() -> None:
    @dataclasses.dataclass
    class Child1:
        a: int = 5

    @dataclasses.dataclass
    class Child2:
        b: int = 3

    @dataclasses.dataclass
    class Parent:
        children: Tuple[Union[Child1, Child2], Union[Child1, Child2]] = (
            tyro.MISSING,
            Child2(2),
        )

    with pytest.raises(SystemExit):
        tyro.cli(Parent, args=[])

    assert tyro.cli(
        Parent,
        args=["children.0:child1", "--children.0.a", "5"],
    ) == Parent((Child1(5), Child2(2)))

    assert tyro.cli(
        Parent,
        args=["children.0:child2", "--children.0.b", "5"],
    ) == Parent((Child2(5), Child2(2)))



from __future__ import annotations

import functools
import sys

from .._docstrings import get_field_docstring
from .._resolver import get_type_hints_resolve_type_params
from .._singleton import MISSING, MISSING_NONPROP
from ._struct_spec import StructConstructorSpec, StructFieldSpec, StructTypeInfo


def msgspec_rule(info: StructTypeInfo) -> StructConstructorSpec | None:
    if "msgspec" not in sys.modules.keys():  # pragma: no cover
        return None

    import msgspec

    try:
        if not issubclass(info.type, msgspec.Struct):
            return None
    except TypeError:  # issubclass failed
        return None

    # Handle msgspec struct objects.
    field_list = []
    struct_type = msgspec.inspect.type_info(info.type)
    assert isinstance(struct_type, msgspec.inspect.StructType)

    # We need to use the original type hints, because `field.type` returns
    # a msgspec-specified type descriptor.
    annotations = get_type_hints_resolve_type_params(info.type, include_extras=True)

    for field in struct_type.fields:
        if info.default not in (
            MISSING,
            MISSING_NONPROP,
        ):
            default = getattr(info.default, field.name)
        elif field.default is not msgspec.NODEFAULT:
            default = field.default
        elif field.default_factory is not msgspec.NODEFAULT:
            default = field.default_factory()
        else:
            default = MISSING_NONPROP

        field_list.append(
            StructFieldSpec(
                name=field.name,
                type=annotations[field.name],
                default=default,
                helptext=functools.partial(
                    get_field_docstring, info.type, field.name, info.markers
                ),
            )
        )

    return StructConstructorSpec(instantiate=info.type, fields=tuple(field_list))



# mypy: ignore-errors
#
# Passing a Union type directly to tyro.cli() doesn't type-check correctly in
# mypy. This will be fixed by `typing.TypeForm`: https://peps.python.org/pep-0747/
"""Nested subcommands

Unions over struct types can be nested to create hierarchical subcommand
structures. By default, nested unions wrapped in ``Annotated`` are flattened
into the parent union. To create a named subcommand group, annotate the nested
union with :func:`tyro.conf.subcommand(name=...)`.

Usage:

    # Show top-level subcommands:
    python ./06_nested_subcommands.py --help

    # The `checkout` subcommand:
    python ./06_nested_subcommands.py checkout --branch main

    # `remote` is a named group containing `push` and `pull`:
    python ./06_nested_subcommands.py remote --help
    python ./06_nested_subcommands.py remote push --remote origin --branch main
    python ./06_nested_subcommands.py remote pull --remote origin
"""

from __future__ import annotations

import dataclasses

from typing_extensions import Annotated

import tyro


@dataclasses.dataclass
class Checkout:
    """Checkout a branch."""

    branch: str


@dataclasses.dataclass
class Commit:
    """Commit changes."""

    message: str


@dataclasses.dataclass
class Push:
    """Push commits to a remote."""

    remote: str = "origin"
    branch: str = "main"


@dataclasses.dataclass
class Pull:
    """Pull commits from a remote."""

    remote: str = "origin"


Remote = Annotated[
    Push | Pull,
    tyro.conf.subcommand(name="remote", description="Remote operations."),
]

if __name__ == "__main__":
    cmd = tyro.cli(Checkout | Commit | Remote)
    print(cmd)



"""Msgspec integration

In addition to standard dataclasses, :func:`tyro.cli()` also supports
`msgspec <https://jcristharif.com/msgspec/>`_ structs.

Usage:

    python ./10_msgspec.py --help
    python ./10_msgspec.py --field1 hello
    python ./10_msgspec.py --field1 hello --field2 5
"""

import msgspec

import tyro


class Args(msgspec.Struct):
    """Description.
    This should show up in the helptext!"""

    field1: str
    """A string field."""

    field2: int = 5
    """A required integer field."""


if __name__ == "__main__":
    args = tyro.cli(Args)
    print(args)



"""Instantiating classes

In addition to functions and dataclasses, we can also generate CLIs from the
constructors of standard Python classes.

Usage:

    python ./13_classes.py --help
    python ./13_classes.py --field1 hello --field2 7
"""

import tyro


class Args:
    def __init__(
        self,
        field1: str,
        field2: int,
        flag: bool = False,
    ):
        """Arguments.

        Args:
            field1: A string field.
            field2: A numeric field.
            flag: A boolean flag.
        """
        self.data = [field1, field2, flag]


if __name__ == "__main__":
    args = tyro.cli(Args)
    print(args.data)



"""Pydantic integration

In addition to standard dataclasses, :func:`tyro.cli()` also supports
`Pydantic <https://github.com/pydantic/pydantic>`_ models.

Usage:

    python ./06_pydantic.py --help
    python ./06_pydantic.py --field1 hello
    python ./06_pydantic.py --field1 hello --field2 5
"""

from pydantic import BaseModel, Field

import tyro


class Args(BaseModel):
    """Description.
    This should show up in the helptext!"""

    field1: str
    field2: int = Field(3, description="An integer field.")


if __name__ == "__main__":
    args = tyro.cli(Args)
    print(args)



from __future__ import annotations

import dataclasses
from typing import Any

from helptext_utils import get_helptext_with_checks

import tyro


def test_recursive_tuple_narrowing() -> None:
    """Test recursive narrowing of nested tuple types."""

    def fn(
        w: Any = (1, 2, 3),
        x: Any = (1, 2, (3, "4", 5)),
        y: tuple[int, int, Any] = (1, 2, (3, 4, 5)),
        z: tuple[int, int, tuple[int, str, int]] = (1, 2, (3, "4", 5)),
    ) -> Any:
        return w, x, y, z

    # Test CLI parsing with nested tuples
    result = tyro.cli(
        fn,
        args="--w 6 7 8 --x 9 10 11 twelve 13 --y.0 14 --y.1 15 --y.2 16 17 18 --z 19 20 21 hello 23".split(),
    )

    helptext = get_helptext_with_checks(fn)
    assert "--w [INT [INT ...]" in helptext
    assert "--x INT INT INT STR INT" in helptext
    assert "--y.0 INT" in helptext
    assert "--y.1 INT" in helptext
    assert "--y.2 [INT [INT ...]]" in helptext
    assert "--z INT INT INT STR INT" in helptext

    # Simple tuple
    assert result[0] == (6, 7, 8)

    # Nested tuple with Any
    assert result[1] == (9, 10, (11, "twelve", 13))
    assert isinstance(result[1][0], int)
    assert isinstance(result[1][1], int)
    assert isinstance(result[1][2], tuple)
    assert isinstance(result[1][2][0], int)
    assert isinstance(result[1][2][1], str)
    assert isinstance(result[1][2][2], int)

    # Tuple with specified types but Any for the nested part
    assert result[2] == (14, 15, (16, 17, 18))
    assert isinstance(result[2][0], int)
    assert isinstance(result[2][1], int)
    assert isinstance(result[2][2], tuple)

    # Tuple with fully specified nested types
    assert result[3] == (19, 20, (21, "hello", 23))
    assert isinstance(result[3][0], int)
    assert isinstance(result[3][1], int)
    assert isinstance(result[3][2], tuple)
    assert isinstance(result[3][2][0], int)
    assert isinstance(result[3][2][1], str)
    assert isinstance(result[3][2][2], int)


def test_tuple_homogeneous_narrowing() -> None:
    """Test narrowing of homogeneous tuples."""

    def fn(
        x: Any = (1, 2, 3),
        y: tuple[Any, ...] = (1, 2, 3),
    ) -> Any:
        return x, y

    result = tyro.cli(fn, args="--x 4 5 6 --y 7 8 9 10".split())

    # Both tuples should be properly narrowed and parsed
    assert result[0] == (4, 5, 6)
    assert result[1] == (7, 8, 9, 10)


def test_mixed_tuple_types() -> None:
    """Test narrowing with mixed tuple element types."""

    def fn(
        x: Any = (1, "two", 3.0),
        y: tuple[int, Any, float] = (1, "two", 3.0),
    ) -> Any:
        return x, y

    result = tyro.cli(fn, args="--x 4 five 6.0 --y 7 eight 9.0".split())

    assert result[0] == (4, "five", 6.0)
    assert isinstance(result[0][0], int)
    assert isinstance(result[0][1], str)
    assert isinstance(result[0][2], float)

    assert result[1] == (7, "eight", 9.0)
    assert isinstance(result[1][0], int)
    assert isinstance(result[1][1], str)
    assert isinstance(result[1][2], float)


def test_partial_tuple_narrowing() -> None:
    """Test that we can narrow tuple types with partially specified types."""

    @dataclasses.dataclass
    class Config:
        x: tuple[int, int, Any] = (1, 2, "hello")
        y: tuple[int, Any, float] = (1, "world", 3.0)

    result = tyro.cli(Config, args="--x 4 5 six --y 7 eight 9.0".split())

    assert result.x == (4, 5, "six")
    assert result.y == (7, "eight", 9.0)
    assert isinstance(result.y[0], int)
    assert isinstance(result.y[1], str)
    assert isinstance(result.y[2], float)



"""Standalone completion script for tyro.

This module contains the completion logic that gets embedded in bash/zsh scripts.
It uses only Python stdlib and is compatible with Python 3.8+.

This file is designed to be:
1. Type-checked by pyright
2. Unit tested directly
3. Embedded as a string in generated completion scripts
"""

import sys
from typing import Any, Dict, List, Tuple


def reconstruct_colon_words(
    words: List[str], current_word_index: int, spec: Dict[str, Any]
) -> Tuple[List[str], int]:
    """Reconstruct words that were split by bash on colons.

    Bash treats ':' as a word separator by default, so typing "dataset:mnist"
    results in words like ["dataset", ":", "mnist"]. This function detects
    such patterns and merges them back together, but only when the merged
    result matches a known subcommand (or is a prefix of one).

    Args:
        words: List of words from the command line (without program name).
        current_word_index: Index of the word being completed (0-based).
        spec: Completion specification dictionary.

    Returns:
        Tuple of (reconstructed_words, new_current_word_index).
    """
    # Get all known subcommands for validation.
    all_subcommands = set(spec.get("subcommands", {}).keys())

    reconstructed = []
    new_cword = current_word_index
    i = 0

    while i < len(words):
        # Pattern 1: word : word (fully typed colon-separated value).
        if (
            i + 2 < len(words)
            and words[i + 1] == ":"
            and not words[i].startswith("-")
            and not words[i + 2].startswith("-")
        ):
            # Tentatively merge.
            merged = words[i] + ":" + words[i + 2]

            # Check if merged matches a known subcommand or is a prefix.
            is_valid = any(
                cmd == merged or cmd.startswith(merged) for cmd in all_subcommands
            )

            if is_valid:
                reconstructed.append(merged)
                # Adjust current word index if we're in or past the merged region.
                if current_word_index >= i + 2:
                    new_cword -= 2
                elif current_word_index == i + 1:
                    # Cursor is on the colon itself.
                    new_cword = len(reconstructed) - 1
                i += 3
                continue

        # Pattern 2: word : (partial typing, colon at end).
        if i + 1 < len(words) and words[i + 1] == ":" and not words[i].startswith("-"):
            merged = words[i] + ":"

            # Check if any subcommand starts with this prefix.
            is_valid = any(cmd.startswith(merged) for cmd in all_subcommands)

            if is_valid:
                reconstructed.append(merged)
                if current_word_index >= i + 1:
                    new_cword -= 1
                elif current_word_index == i:
                    new_cword = len(reconstructed) - 1
                i += 2
                continue

        # Pattern 3: : word (typing after colon).
        if (
            i > 0
            and words[i] == ":"
            and i + 1 < len(words)
            and not words[i + 1].startswith("-")
        ):
            # Check if previous word exists and could form a subcommand.
            if reconstructed and not reconstructed[-1].startswith("-"):
                merged = reconstructed[-1] + ":" + words[i + 1]
                is_valid = any(
                    cmd == merged or cmd.startswith(merged) for cmd in all_subcommands
                )

                if is_valid:
                    # Replace the last reconstructed word with the merged version.
                    reconstructed[-1] = merged
                    if current_word_index >= i + 1:
                        new_cword -= 1
                    i += 2
                    continue

        # No merge pattern matched.
        if words[i] != ":":  # Skip standalone colons that weren't merged.
            reconstructed.append(words[i])
            i += 1
        else:
            # Standalone colon that didn't match any pattern.
            if current_word_index > i:
                new_cword -= 1
            i += 1

    return reconstructed, new_cword


def get_completions(
    words: List[str], current_word_index: int, spec: Dict[str, Any]
) -> List[str]:
    """Get completions for the current position.

    Args:
        words: List of words on the command line (including the program name).
        current_word_index: Index of the word being completed (0-based).
        spec: Completion specification dictionary.

    Returns:
        List of completion strings in "completion\tdescription" format.
    """
    # Skip the program name.
    if current_word_index == 0:
        return []

    words = words[1:]  # Remove program name.
    current_word_index -= 1

    # Reconstruct words that were split by bash on colons.
    words, current_word_index = reconstruct_colon_words(words, current_word_index, spec)

    # Get the word being completed.
    current_word = words[current_word_index] if current_word_index < len(words) else ""

    # Helper function to navigate to current context.
    def navigate_to_context(end_index: int) -> Dict[str, Any]:
        """Navigate through subcommands and options to find current context."""
        ctx = spec
        i = 0
        while i < end_index:
            word = words[i]
            if word in ctx.get("subcommands", {}):
                ctx = ctx["subcommands"][word]
                i += 1
            elif word.startswith("-"):
                # Skip options and their arguments.
                takes_arg = False
                for opt in ctx.get("options", []):
                    if word in opt["flags"]:
                        if opt["type"] not in ("flag", "boolean"):
                            takes_arg = True
                        break
                i += 1
                if takes_arg and i < end_index:
                    i += 1  # Skip the argument value.
            else:
                i += 1
        return ctx

    # Check if we're completing an option argument (like after --mode).
    # This includes nargs>1 cases like: --modes train <TAB>
    if current_word_index > 0:
        # Navigate to current context.
        current_spec = navigate_to_context(current_word_index)

        # Look backwards to find if we're filling an option argument.
        for j in range(current_word_index - 1, -1, -1):
            prev = words[j]
            if prev.startswith("-"):
                # Found an option flag.
                for opt in current_spec.get("options", []):
                    if prev in opt["flags"]:
                        # Check if this option takes arguments and has choices.
                        if opt["type"] == "choice" and "choices" in opt:
                            # Check if we should still be completing values for this option.
                            # This handles nargs > 1 (e.g., nargs=2, nargs='+', nargs='*').
                            nargs = opt.get("nargs")
                            if nargs in ("+", "*") or (
                                isinstance(nargs, int) and nargs > 1
                            ):
                                # Multi-value option: keep offering choices.
                                # Count how many values we've already provided.
                                values_provided = current_word_index - j - 1
                                if isinstance(nargs, int):
                                    # Stop if we've provided enough values.
                                    if values_provided >= nargs:
                                        break
                                # Continue offering choices.
                                completions = []
                                for choice in opt["choices"]:
                                    if str(choice).startswith(current_word):
                                        completions.append(f"{choice}\t{choice}")
                                return completions
                            elif j == current_word_index - 1:
                                # Single-value option: only complete if immediately after flag.
                                completions = []
                                for choice in opt["choices"]:
                                    if str(choice).startswith(current_word):
                                        completions.append(f"{choice}\t{choice}")
                                return completions
                        elif (
                            opt["type"] not in ("flag", "boolean")
                            and j == current_word_index - 1
                        ):
                            # Check if it's a path type - return marker for shell.
                            if opt["type"] == "path":
                                return ["__TYRO_COMPLETE_FILES__"]
                            # For other option types, don't provide completions.
                            return []
                # We found a flag, stop searching.
                break

    # Navigate to the current subcommand context.
    # Keep track of root spec and selected frontier subcommands.
    root_spec = spec
    current_spec = spec
    used_words = set()
    selected_frontier_subcommands = []
    has_frontier = len(spec.get("frontier_groups", [])) > 0
    i = 0

    while i < current_word_index:
        word = words[i]
        used_words.add(word)

        # Check if this word is a subcommand.
        if word in current_spec.get("subcommands", {}):
            # In frontier mode, don't navigate into subcommands - stay at root.
            # Instead, track selected frontier subcommands for cascade options.
            if has_frontier:
                selected_frontier_subcommands.append(word)
                i += 1
            else:
                # Normal mode: navigate into the subcommand.
                current_spec = current_spec["subcommands"][word]
                i += 1
        elif word.startswith("-"):
            # Skip options and their arguments.
            # Find the option definition to see if it takes an argument.
            takes_arg = False
            for opt in current_spec.get("options", []):
                if word in opt["flags"]:
                    if opt["type"] not in ("flag", "boolean"):
                        takes_arg = True
                    break

            i += 1
            if takes_arg and i < current_word_index:
                i += 1  # Skip the argument value.
        else:
            i += 1

    # Generate completions.
    completions = []

    # Add available options from current spec.
    for opt in current_spec.get("options", []):
        for flag in opt["flags"]:
            if flag.startswith(current_word):
                # Format: flag\tdescription (tab-separated).
                completions.append(f"{flag}\t{opt['description']}")

    # In frontier mode, also add cascade options from selected subcommands.
    if has_frontier and selected_frontier_subcommands:
        for subcmd in selected_frontier_subcommands:
            subcmd_spec = root_spec["subcommands"].get(subcmd, {})
            for opt in subcmd_spec.get("options", []):
                # Only include options marked as cascade.
                if opt.get("cascade", False):
                    for flag in opt["flags"]:
                        if flag.startswith(current_word):
                            completions.append(f"{flag}\t{opt['description']}")

    # Add available subcommands.
    # Only use root frontier groups if we're still at root level.
    frontier_groups = current_spec.get("frontier_groups", [])
    if not frontier_groups and current_spec == root_spec:
        frontier_groups = root_spec.get("frontier_groups", [])

    if frontier_groups:
        # Frontier mode: offer subcommands from groups that haven't been used yet.
        for group in frontier_groups:
            # Check if any subcommand from this group has been used.
            group_used = any(cmd in used_words for cmd in group)
            if not group_used:
                # Offer all subcommands from this group.
                for cmd in group:
                    if cmd.startswith(current_word):
                        # Look up description from root spec.
                        desc = (
                            root_spec["subcommands"]
                            .get(cmd, {})
                            .get("description", cmd)
                        )
                        completions.append(f"{cmd}\t{desc}")
    else:
        # Normal mode: offer all subcommands.
        for cmd, cmd_spec in current_spec.get("subcommands", {}).items():
            if cmd.startswith(current_word):
                completions.append(f"{cmd}\t{cmd_spec['description']}")

    return completions


def main() -> None:
    """Main entrypoint when run as a script.

    Expects command line arguments in the format:
    <word1> <word2> ... <wordN> <current_index>

    The COMPLETION_SPEC variable should be defined before calling this.
    """
    if len(sys.argv) < 2:
        sys.exit(0)

    current_index = int(sys.argv[-1])
    words = sys.argv[1:-1]

    # COMPLETION_SPEC is expected to be defined in the embedding context.
    completions = get_completions(words, current_index, COMPLETION_SPEC)  # type: ignore[name-defined]  # noqa: F821

    # Output one completion per line.
    for completion in completions:
        print(completion)


if __name__ == "__main__":
    main()



"""tyro's API relies heavily on parsing type annotations, which may be both
"regular" types like `int`, `str`, and classes, or more general forms like `int
| str` or `Annotated[str, ...]`.

To correctly annotate variables that can take these more general type forms, we
need to wait for the typing.TypeForm PEP:

    https://peps.python.org/pep-0747/

This is still in draft form, so in the meantime we use Type[T] everywhere we would
otherwise have TypeForm[T]. This mostly works, and fortunately is supported by
pyright and pylance (relevant:
https://github.com/microsoft/pyright/issues/4298), but should be switched for
the correct typing.TypeForm annotation once it's available.

Also relevant:
- mypy support for typing_extensions.TypeForm: https://github.com/python/mypy/issues/9773
"""

from typing import Type as TypeForm

__all__ = ["TypeForm"]
